#!/usr/bin/env python
# -*- coding: utf-8 -*-

#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>>Custom Font Tools>>Custom Font Layers From Fonts With Forced Margins vx.xx...
#
# Changelog
# Version 0.01 Options for creating sets of custom font character layers with specified margins and scaled to size
# Version 0.02 Added options for creating limited sets of unicode character layers
# Version 0.03 Implemented greater range of unicode characters
# Version 0.04 Increased initial image size to accomodate longer descender problem, fixed one bug, added tabbed dialog, added Dutch/Portuguese accented characters


#from gimpfu import *
import unicodedata
import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################

#from array import array
#import math
#import random

#from gtk import gdk
#w = gdk.screen_width()
#h = gdk.screen_height()

# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)


def do_stuff(image, layer, font_name, font_size, left_margin, top_margin, right_margin, bottom_margin, add_upper_case, add_lower_case, add_numbers, add_punctuation, add_others, add_specials, restricted_char_set, restricted_case, language_ptr, symbols, crop_symbols, name_symbols, display_keywords, scale_custom, scale_to_pixel, scale_size, ptab):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
	
    dummy = 0
    image.active_layer =  layer
	
    do_job = False
	
    add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, False, False, False, False
	
    if display_keywords:
        message_text = "Current keywords for grouping unicode characters:\n\nactivity animal arrow asterisk astrology \nballot beauty body braille bubble building bullet \ncandy celebration cereal chess circle clock clothes clubs communication cross "
        message_text += "currency \ndiamonds dice dingbat domino \neating emoticon entertainment \nface fantasy figures finance flag flower food fruit \ngender gestures hands health \nhearts \nmap media medicine misc money music \nnumber \npen people places pointer \nrectangle recycling rhombus "
        message_text += "religion romance runic \nscissors shape sign space spades sport square star stationery sweets suit \ntarot technology tick transport tree triangle \nvegetable vote \nweather."
        old_handler = pdb.gimp_message_get_handler()	
        pdb.gimp_message_set_handler(MESSAGE_BOX)
        pdb.gimp_message(message_text)
        pdb.gimp_message_set_handler(old_handler)
        
    # check for unwanted space characters in input string
    restricted_case =  restricted_case.strip()
    if restricted_case == "": restricted_case = "321cbaCBA"
    visible_layer_name = "undefined"
    if " " in restricted_case:
        loop, temp = 0, ""
        while loop < len(restricted_case):
            char = restricted_case[loop]
            if char <> " ": temp += char
            loop += 1
        if temp <> "":
            restricted_case = temp
        else:
           restricted_case = "321gfeCBA"
		   
    symbols = symbols.lower() # ensure that later tests only need lower case
		   
    if restricted_char_set:
        restricted_case = restricted_case.decode("utf-8")
        layer_string = restricted_case
        symbol_list = ["dummy"]
        do_job = True
            
    else:
        # characters in reverse order since the code works form bottom to top layer
        upper_case = "ZYXWVUTSRQPONMLKJIHGFEDCBA" # "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
        lower_case = "zyxwvutsrqponmlkjihgfedcba" #  "abcdefghijklmnopqrstuvwxyz"
        number_case = "0987654321" # "1234567890"
        punctuation_case = unichr(34) + "!;',.:?" # character 34 is the double quotes character
		
        dutch_case = unichr(192) + unichr(193)  + unichr(194) + unichr(196) +  unichr(200) + unichr(201) + unichr(202)  + unichr(203) # A grave + A acute + A circumflex + A diaeresis + E grave + E acute  + E circumflex + E diaeresis
        dutch_case += unichr(204) + unichr(205) + unichr(206) + unichr(207) + unichr(306)  + unichr(210) + unichr(211) + unichr(212) + unichr(214)# I grave + I acute + I circumflex + I diaeresis + ligature IJ + O grave + O acute +  O circumflex + O  diaeresis
        dutch_case += unichr(217) + unichr(218) + unichr(219) + unichr(220)  + unichr(221) + unichr(376) # U grave + U acute + U circumflex + U diaeresis + Y acute + Y diaeresis
        dutch_case += unichr(224) + unichr(225) + unichr(226) + unichr(228) + unichr(232) + unichr(233) + unichr(234) + unichr(235) #  a grave + a acute + a circumflex + a diaeresis  + e grave + e acute + e circumflex + e diaeresis
        dutch_case += unichr(236) + unichr(237) + unichr(238) + unichr(239) + unichr(307) + unichr(242) + unichr(244) + unichr(246) # i grave + i acute + i circumflex + i diaeresis + ligature ij + o grave + o circumflex + o  diaeresis
        dutch_case += unichr(249) + unichr(251) + unichr(252) + unichr(253) + unichr(255) # u grave + u circumflex  + u diaeresis + y acute + y diaeresis 
		
        french_case = unichr(192) + unichr(194) + unichr(196) + unichr(198) + unichr(199) # A grave + A circumflex + A diaeresis + AE ligature + C cedilla
        french_case +=  unichr(200) + unichr(201) + unichr(202) + unichr(203) + unichr(206) + unichr(207) # E grave + E acute + E circumflex + E diaeresis + I circumflex + I diaeresis
        french_case += unichr(212) + unichr(217) + unichr(219) + unichr(220) # O circumflex + U grave + U circumflex + U diaeresis
        french_case += unichr(224) + unichr(226) + unichr(228) + unichr(230) + unichr(231) # a grave + a circumflex + a diaeresis + ae ligature + c cedilla
        french_case += unichr(232) + unichr(233) + unichr(234) + unichr(235) + unichr(238) + unichr(239) # e grave + e acute + e circumflex + e diaeresis + i circumflex + i diaeresis
        french_case += unichr(244) + unichr(249) + unichr(251) + unichr(252) + unichr(338) + unichr(339) + unichr(376) + unichr(255) # o circumflex + u grave + u circumflex + u diaeresis + OE ligature + oe ligature + Y diaeresis + y diaeresis
		
        german_case = unichr(196) + unichr(214) + unichr(220) + unichr(223) # A diaeresis + O diaeresis + U diaeresis + Eszett
        german_case += unichr(228) + unichr(246) + unichr(252) # a diaeresis + o diaeresis + u diaeresis
		
        italian_case = unichr(192) + unichr(200) + unichr(201) + unichr(204) + unichr(210) + unichr(217) #  A grave + E grave + E acute + I grave + O grave + U grave
        italian_case += unichr(224) + unichr(232) + unichr(233) + unichr(236) + unichr(242) + unichr(249) # a grave + e grave + e acute + i grave + o grave + u grave
		
        polish_case = unichr(260) + unichr(262) + unichr(280) + unichr(321) + unichr(323) + unichr(211) + unichr(346) + unichr(377) + unichr(379) # A ogonek + C acute + E ogonek + L stroke + N acute + O acute + S acute + Z acute + Z dot above
        polish_case += unichr(261) + unichr(263) + unichr(281) + unichr(322) + unichr(324) + unichr(243) + unichr(347) + unichr(378) + unichr(380) # a ogonek + c acute + e ogonek + l stroke + n acute + o acute + s acute + z acute + z dot above
		
        portuguese_case =  unichr(192) + unichr(193) + unichr(194) + unichr(195) + unichr(201) + unichr(202) + unichr(205) + unichr(211) + unichr(212) + unichr(213) # A grave + A acute + A circumflex + A tilde + E acute + E circumflex + I acute + O acute + O circumflex + O tilde
        portuguese_case += unichr(218) + unichr(199) + unichr(220) + unichr(224) + unichr(225) + unichr(226) + unichr(227) + unichr(233) + unichr(234) # U Acute + C cedilla + U diaeresis + a grave + a acute + a circumflex + a tilde + e acute + e circumflex
        portuguese_case += unichr(237) + unichr(243) + unichr(244) + unichr(245) + unichr(250)  + unichr(231) + unichr(252) # i acute + o acute + o circumflex + o tilde + u acute + c cedilla + u diaeresis
		
        spanish_case = unichr(161) + unichr(191) + unichr(193) + unichr(201) + unichr(205) + unichr(209) + unichr(211) # inverted exclamation + inverted question + A acute + E acute + I acute + N tilde + O acute
        spanish_case += unichr(218) + unichr(220) + unichr(225) + unichr(233) + unichr(237) + unichr(241) # U Acute + U diaeresis + a acute + e acute + i acute + n tilde
        spanish_case += unichr(243) + unichr(250) + unichr(252)# o acute + u acute + u diaeresis
		
        other_case = "-=`[]\/~@#$" + unichr(163) + "%^&*()_+{}|<>" # other characters from the keyboard - character 163 is the pound sterling sign
		
        special_case = unichr(169) + unichr(174) + unichr(8364) + unichr(188) + unichr(189) + unichr(190) + unichr(162) # Copyright, Registered Trademark, Euro sign, one quarter sign, one half sign, three quarters sign, cent sign
        special_case =  special_case + unichr(176) + unichr(185) + unichr(178) + unichr(179) + unichr(183) + unichr(215) + unichr(247) # degree sign, superscript 1, superscript 2, superscript 3, middle dot, multiplication, division
        special_case =  special_case + unichr(8220) + unichr(8216) # smart right hand double quote + smart right hand single quote

        symbol_case = ""
        
        symbol_list = ["dummy"]

        #### start of unicode character data 
			
        if "anim" in symbols:# animals 
            do_job = True
            #symbol_list = fn_get_unicode_value(128000, 128063,symbol_list)
            symbol_list = fn_get_unicode_value(128049, 128062,symbol_list)
            symbol_list = fn_get_unicode_value(129408, 129431,symbol_list)
						
        if "arro" in symbols or "arrow plain" in symbols:# arrows
            do_job = True
            symbol_list = fn_get_unicode_value(129024,129095,symbol_list)# from supplementary arrows C 
            symbol_list = fn_get_unicode_value(129104,129113,symbol_list)# from supplementary arrows C 
            symbol_list = fn_get_unicode_value(129120,129159,symbol_list)# from supplementary arrows C 
            symbol_list = fn_get_unicode_value(129168,129179,symbol_list)# from supplementary arrows C 
			
        if "arro" in symbols or "arrow 3d" in symbols:# arrows
            do_job = True
            symbol_list = fn_get_unicode_value(129184,129195,symbol_list)# from supplementary arrows C 
			
        if "arro" in symbols or "arrow curved" in symbols:# arrows
            do_job = True
            symbol_list = fn_get_unicode_value(10552,10561,symbol_list)# from supplementary arrows B
            symbol_list = fn_get_unicode_value(10226,10227,symbol_list)# from supplementary arrows A
				
        if "aste" in symbols:# asterisks
            do_job = True
            symbol_list = fn_get_unicode_value(10018,10021,symbol_list)
            symbol_list = fn_get_unicode_value(10031,10033,symbol_list)
            symbol_list = fn_get_unicode_value(10040,10043,symbol_list)
            symbol_list = fn_get_unicode_value(10048,0,symbol_list)
            symbol_list = fn_get_unicode_value(10052,10057,symbol_list)
            symbol_list = fn_get_unicode_value(10059,0,symbol_list)
			
				
        if "astrol" in symbols or "cons" in symbols:# astrology constellations
            do_job = True
            symbol_list = fn_get_unicode_value(9800,9811,symbol_list)
				
        if "beaut" in symbols:# beauty
            do_job = True
            symbol_list = fn_get_unicode_value(128132,128136,symbol_list)
				
        if "body" in symbols:# 
            do_job = True
            symbol_list = fn_get_unicode_value(128064,128080,symbol_list)
				
        if "brai" in symbols:# braille
            do_job = True
            symbol_list = fn_get_unicode_value(10241,10495,symbol_list)
				
        if "bubb" in symbols or "speech" in symbols:# speech bubbles
            do_job = True
            symbol_list = fn_get_unicode_value(128172,128173,symbol_list)
            symbol_list = fn_get_unicode_value(128488,128498,symbol_list)
				
        if "buil" in symbols:# buildings
            do_job = True
            symbol_list = fn_get_unicode_value(127967,127984,symbol_list)
				
        if "bull" in symbols:# bullet symbols
            do_job = True
            symbol_list = fn_get_unicode_value(9152,0,symbol_list)
            symbol_list = fn_get_unicode_value(10070,0,symbol_list)
            symbol_list = fn_get_unicode_value(10084,10087,symbol_list)
            symbol_list = fn_get_unicode_value(10102,10131,symbol_list)
				
        if "busi" in symbols:# buildings
            do_job = True
            symbol_list = fn_get_unicode_value(128186,128230,symbol_list)
	
        if "cards spades" in symbols or "spades" in symbols:# playing cards spades
            do_job = True
            symbol_list = fn_get_unicode_value(127137,127150,symbol_list)
				
        if "cards hearts" in symbols or "hearts" in symbols:# playing cards hearts
            do_job = True
            symbol_list = fn_get_unicode_value(127153,127166,symbol_list)

        if "cards diamonds" in symbols or "diamonds" in symbols:# playing cards diamonds
            do_job = True
            symbol_list = fn_get_unicode_value(127169,127182,symbol_list)

        if "cards clubs" in symbols or "clubs" in symbols:# playing cards clubs
            do_job = True
            symbol_list = fn_get_unicode_value(127185,127198,symbol_list)

        if "card suit" in symbols or "suits" in symbols:# playing cards the four suit symbols
            do_job = True
            symbol_list = fn_get_unicode_value(9824,0,symbol_list)
            symbol_list = fn_get_unicode_value(9827,0,symbol_list)
            symbol_list = fn_get_unicode_value(9829,0,symbol_list)
            symbol_list = fn_get_unicode_value(9830,0,symbol_list)
			
        if "cele" in symbols: # celebrations
            do_job = True
            symbol_list = fn_get_unicode_value(127872,127893,symbol_list)
			
        if "cer" in symbols: # cereals
            do_job = True
            symbol_list = fn_get_unicode_value(127805,127806,symbol_list)
			
        if "chess" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9812,9823,symbol_list)
			
        if "circ" in symbols:# Geometric Shapes 
            do_job = True
            symbol_list = fn_get_unicode_value(9673,0,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9675,9687,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9711,0,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9716,9719,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(128900,128907,symbol_list) # Geometric Shapes Extended
			
        if "clock" in symbols or "time" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9200,9203,symbol_list)
            symbol_list = fn_get_unicode_value(128336,128359,symbol_list)
			
        if "cloth" in symbols: # clothes
            do_job = True
            symbol_list = fn_get_unicode_value(128081,128098,symbol_list)
			
        if "comm" in symbols: # communication
            do_job = True
            symbol_list = fn_get_unicode_value(128379,128390,symbol_list)

        if "cross" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(10009,10016,symbol_list)

        if "dice" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9856,9861,symbol_list)
				
        if "dingbat" in symbols or "ding" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9984,10175,symbol_list)
								
        if "domino" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(127024,127123,symbol_list)
								
        if "ente" in symbols: # entertainment
            do_job = True
            symbol_list = fn_get_unicode_value(127896,127932,symbol_list)
				
        if "face" in symbols or "emoti" in symbols : # emoticons
            do_job = True
            symbol_list = fn_get_unicode_value(9785,9787,symbol_list)
            symbol_list = fn_get_unicode_value(128512,128591,symbol_list)
            symbol_list = fn_get_unicode_value(129296,129303,symbol_list)
            symbol_list = fn_get_unicode_value(129312,129327,symbol_list)
            symbol_list = fn_get_unicode_value(129488,0,symbol_list)
				
        if "fant" in symbols: # fantasy
            do_job = True
            symbol_list = fn_get_unicode_value(128120,128128,symbol_list)
				
        if "figures" in symbols or "people" in symbols:# people
            do_job = True
            symbol_list = fn_get_unicode_value(129328,129329,symbol_list)
            symbol_list = fn_get_unicode_value(129332,129335,symbol_list)
            symbol_list = fn_get_unicode_value(129489,129503,symbol_list)
				
        if "flag" in symbols:# flags
            do_job = True
            symbol_list = fn_get_unicode_value(127985,127988,symbol_list)
				
        if "flow" in symbols:# flowers
            do_job = True
            symbol_list = fn_get_unicode_value(127799,127804,symbol_list)
				
        if "food" in symbols or "eating" in symbols:# food and eating 
            do_job = True
            symbol_list = fn_get_unicode_value(129360,129387,symbol_list)
            symbol_list = fn_get_unicode_value(127789,127791,symbol_list)
            symbol_list = fn_get_unicode_value(127798,0,symbol_list)
            symbol_list = fn_get_unicode_value(127807,0,symbol_list)
            symbol_list = fn_get_unicode_value(127828,127850,symbol_list)
            symbol_list = fn_get_unicode_value(127854,127859,symbol_list)
            symbol_list = fn_get_unicode_value(127860,127870,symbol_list)
			
        if "fru" in symbols: # fruit
            do_job = True
            symbol_list = fn_get_unicode_value(127813,0,symbol_list)
            symbol_list = fn_get_unicode_value(127815,127827,symbol_list)
				
        if "gend" in symbols:# gender
            do_job = True
            symbol_list = fn_get_unicode_value(9890,9895,symbol_list)
			
        if "hand" in symbols or "gest" in symbols:# hands
            do_job = True
            symbol_list = fn_get_unicode_value(128398,128419,symbol_list)
            symbol_list = fn_get_unicode_value(129304,129311,symbol_list)
            symbol_list = fn_get_unicode_value(129330,129331,symbol_list)
			
        if "heal" in symbols or "medi" in symbols:# health medicine
            do_job = True
            symbol_list = fn_get_unicode_value(128137,128138,symbol_list)
			
        if "media" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(128231,128280,symbol_list)
            symbol_list = fn_get_unicode_value(128360,128362,symbol_list)
			
        if "misc" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(8488,8527,symbol_list)
				
        if "money" in symbols or "curr" in symbols or "fina" in symbols:# money currency finance 
            do_job = True
            symbol_list = fn_get_unicode_value(8352,8383,symbol_list)
            symbol_list = fn_get_unicode_value(128176,128185,symbol_list)
				
        if "music" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9833,9839,symbol_list)
            symbol_list = fn_get_unicode_value(119040,119078,symbol_list)
            symbol_list = fn_get_unicode_value(119081,119106,symbol_list)
            symbol_list = fn_get_unicode_value(119130,119140,symbol_list)
			
        if "number" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(8528,8587,symbol_list)
			
        if "pen" in symbols: #pen pencil
            do_job = True
            symbol_list = fn_get_unicode_value(9998,10002,symbol_list)
			
        if "peo" in symbols: # people
            do_job = True
            symbol_list = fn_get_unicode_value(128100,128120,symbol_list)
            symbol_list = fn_get_unicode_value(128129,128131,symbol_list)
				
        if "place" in symbols:# places
            do_job = True
            symbol_list = fn_get_unicode_value(127956,127966,symbol_list)
				
        if "point" in symbols:# pointers
            do_job = True
            symbol_list = fn_get_unicode_value(9754,9759,symbol_list)
				
        if "rect" in symbols:# Geometric Shapes 
            do_job = True
            symbol_list = fn_get_unicode_value(9644,9647,symbol_list) # Geometric Shapes
			
        if "recy" in symbols:# recycling symbols
            do_job = True
            symbol_list = fn_get_unicode_value(9842,9854,symbol_list)
			
        if "rhom" in symbols:# Geometric Shapes 
            do_job = True
            symbol_list = fn_get_unicode_value(9670,9672,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9674,0,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(128919,128928,symbol_list) # Geometric Shapes Extended
			
        if "romanc" in symbols or "wedd" in symbols:# romance wedding
            do_job = True
            symbol_list = fn_get_unicode_value(128139,128160,symbol_list)

        if "relig" in symbols or "god" in symbols:# religion
            do_job = True
            symbol_list = fn_get_unicode_value(9765,9775,symbol_list)
            symbol_list = fn_get_unicode_value(9784,0,symbol_list)
            symbol_list = fn_get_unicode_value(128325,128334,symbol_list)

        if "runic" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(5792,5872,symbol_list)
				
        if "scis" in symbols:# scissors
            do_job = True
            symbol_list = fn_get_unicode_value(9984,9988,symbol_list)
			
        if "shape" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9632,9727,symbol_list)
            symbol_list = fn_get_unicode_value(11039,11055,symbol_list)
			
        if "sign" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9760,9764,symbol_list)
            symbol_list = fn_get_unicode_value(9773,9774,symbol_list)
            symbol_list = fn_get_unicode_value(9874,9884,symbol_list)
            symbol_list = fn_get_unicode_value(9888,0,symbol_list)
            symbol_list = fn_get_unicode_value(9935,9939,symbol_list)

        if "space" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9788,9799,symbol_list)
            symbol_list = fn_get_unicode_value(9732,0,symbol_list)

        if "sport" in symbols or "activity" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9917,9918,symbol_list)
            symbol_list = fn_get_unicode_value(9975,9977,symbol_list)
            symbol_list = fn_get_unicode_value(127933,127955,symbol_list)
            symbol_list = fn_get_unicode_value(127992,127993,symbol_list)
            symbol_list = fn_get_unicode_value(129336,129342,symbol_list)
            symbol_list = fn_get_unicode_value(129349,129356,symbol_list)
			
        if "squar" in symbols:# Geometric Shapes 
            do_job = True
            symbol_list = fn_get_unicode_value(9632,9643,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9703,9707,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9712,9715,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9723,9726,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(128908,128918,symbol_list) # Geometric Shapes Extended
				
        if "star" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9956,9959,symbol_list)
            symbol_list = fn_get_unicode_value(10017,0,symbol_list) # Dingbats
            symbol_list = fn_get_unicode_value(10022,10032,symbol_list) # Dingbats
            symbol_list = fn_get_unicode_value(10036,10041,symbol_list) # Dingbats
            symbol_list = fn_get_unicode_value(128960,128980,symbol_list) # Geometric Shapes Extended
				
        if "stat" in symbols: # stationery
            do_job = True
            symbol_list = fn_get_unicode_value(128391,128397,symbol_list)
				
        if "swe" in symbols or "cand" in symbols or "conf" in symbols: # sweets candy confectionary
            do_job = True
            symbol_list = fn_get_unicode_value(127851,127853,symbol_list)
            symbol_list = fn_get_unicode_value(127871,0,symbol_list)
				
        if "tarot" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(127200,127221,symbol_list)
				
        if "tech" in symbols:# technology 
            do_job = True
            symbol_list = fn_get_unicode_value(9167,0,symbol_list)
            symbol_list = fn_get_unicode_value(9193,9199,symbol_list)
            symbol_list = fn_get_unicode_value(9204,9214,symbol_list)
            symbol_list = fn_get_unicode_value(128421,128443,symbol_list)
				
        if "tick" in symbols or "check" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(10003,10004,symbol_list)
			
        if "transp" in symbols or "map" in symbols:# transport/map symbols
            do_job = True
            symbol_list = fn_get_unicode_value(128640,128722,symbol_list)
            symbol_list = fn_get_unicode_value(128725,0,symbol_list)
            symbol_list = fn_get_unicode_value(128736,128748,symbol_list)
            symbol_list = fn_get_unicode_value(128752,128762,symbol_list)
			
        if "tree" in symbols: 
            do_job = True
            symbol_list = fn_get_unicode_value(127792,127796,symbol_list)
            symbol_list = fn_get_unicode_value(127809,127811,symbol_list)
			
        if "triang" in symbols:# Geometric Shapes 
            do_job = True
            symbol_list = fn_get_unicode_value(9650,9669,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9698,9701,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9708,9710,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9720,9722,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(9727,0,symbol_list) # Geometric Shapes
            symbol_list = fn_get_unicode_value(128896,128899,symbol_list) # Geometric Shapes Extended
			
        if "veg" in symbols:# vegetables
            do_job = True
            symbol_list = fn_get_unicode_value(127812,127814,symbol_list)
				
        if "vote" in symbols or "ballot" in symbols:
            do_job = True
            symbol_list = fn_get_unicode_value(9745,9746,symbol_list)
            symbol_list = fn_get_unicode_value(10007,10008,symbol_list)

        if "weather" in symbols: 
            do_job = True
            symbol_list = fn_get_unicode_value(9728,9737,symbol_list)
            symbol_list = fn_get_unicode_value(9924,9928,symbol_list)
            symbol_list = fn_get_unicode_value(127744,127755,symbol_list)
            symbol_list = fn_get_unicode_value(127777,127788,symbol_list)
			
        #### end of unicode character data
			
        layer_string = "" # Begin adding characters to layer_string
		
        # select non-English language accented characters to add
		
        if language_ptr == 0: #None
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, False, False, False, False
        elif language_ptr == 1: # Dutch
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = True, False, False, False, False, False, False
        elif language_ptr == 2: #French
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, True, False, False, False, False, False
        elif language_ptr == 3:  #German
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, True, False, False, False, False
        elif language_ptr == 4: #Italian
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, True, False, False, False
        elif language_ptr == 5: #Polish
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, False, True, False, False
        elif language_ptr == 6: #Portuguese
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, False, False, True, False
        elif language_ptr == 7: #Spanish
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = False, False, False, False, False, False, True
        else: # all
            add_dutch, add_french, add_german, add_italian, add_polish, add_portuguese, add_spanish = True, True, True, True, True, True, True
		
            
        if add_dutch:
            do_job = True
            for loop in range(len(dutch_case)):
                if dutch_case[loop] not in layer_string:
                    layer_string = dutch_case[loop] + layer_string
		
        if add_french:
            do_job = True
            for loop in range(len(french_case)):
                if french_case[loop] not in layer_string:
                    layer_string = french_case[loop] + layer_string
					
        if add_german:
            do_job = True
            for loop in range(len(german_case)):
                if german_case[loop] not in layer_string:
                    layer_string = german_case[loop] + layer_string
					
        if add_italian:
            do_job = True
            for loop in range(len(italian_case)):
                if italian_case[loop] not in layer_string:
                    layer_string = italian_case[loop] + layer_string
					
        if add_polish:
            do_job = True
            for loop in range(len(polish_case)):
                if polish_case[loop] not in layer_string:
                    layer_string = polish_case[loop] + layer_string
					
        if add_portuguese:
            do_job = True
            for loop in range(len(portuguese_case)):
                if portuguese_case[loop] not in layer_string:
                    layer_string = portuguese_case[loop] + layer_string
					
        if add_spanish:
            do_job = True
            for loop in range(len(spanish_case)):
                if spanish_case[loop] not in layer_string:
                    layer_string = spanish_case[loop] + layer_string
					
        layer_string += symbol_case
		
        if add_specials: 
            do_job = True
            layer_string += special_case

        if add_others: 
            do_job = True
            layer_string += other_case
		
        if add_punctuation: 
            do_job = True
            layer_string += punctuation_case
		
        if add_numbers: 
            do_job = True
            layer_string += number_case
		
        if add_lower_case: 
            do_job = True
            layer_string += lower_case
		
        if add_upper_case: 
            do_job = True
            layer_string += upper_case
		
		
    if do_job == True:
            
        # remove existing empty layers
        layer_names_list = image.layers
        loop = 0
        while loop<len(layer_names_list):
            pdb.gimp_image_select_item(image, 2, layer_names_list[loop])
            empty = pdb.gimp_selection_is_empty(image)
            pdb.gimp_selection_none(image)
            if empty: pdb.gimp_image_remove_layer(image, layer_names_list[loop])
            loop += 1

        min_y_pos = 1000000
        min_x_pos = 1000000
        max_y_pos = 0
        max_x_pos = 0
        removed = "None"
	
        pdb.gimp_image_resize(image, font_size*3, font_size*3,0,0)
        loop = 0
        #pdb.gimp_message("len layer_string = " + str(len(layer_string)) + " " + layer_string + " Len symbol_list = " + str(len(symbol_list)))
		
        if len(symbol_list)>1:
            loop = 1
            while loop < len(symbol_list): # loop through unicode list to find the max/min positions of characters in all layers
                text_layer = pdb.gimp_text_fontname(image, None,0, 0, eval(symbol_list[loop]), 0, 1, font_size, 1, "dummy")
                pdb.gimp_image_select_item(image, 2, text_layer) # character alpha to selection 
                selection,x0,x1,y0,y1, s_width, s_height = fn_selection_info(image) # get alpha selection info
                pdb.gimp_selection_none(image)
                if selection:
                    if x0<min_x_pos: min_x_pos = x0 # find the minimum x co-ord of all characters
                    if y0<min_y_pos: min_y_pos = y0 # find the minimum y co-ord of all characters
                    if x1>max_x_pos: max_x_pos = x1 # find the maximum x co-ord of all characters
                    if y1>max_y_pos: max_y_pos = y1 # find the maximum y co-ord of all characters
			
                pdb.gimp_image_remove_layer(image, text_layer) # discard each layer as we go along after gathering data
                loop += 1
        else:
		
            while loop < len(layer_string): # loop through characters to find the max/min positions of characters in all layers
                #char = layer_string[loop]
                char = layer_string[loop].decode('utf-8')
                if char <> " ":
                    text_layer = pdb.gimp_text_fontname(image, None,0, 0, char, 0, 1, font_size, 1, font_name)
                    pdb.gimp_image_select_item(image, 2, text_layer) # character alpha to selection 
                    selection,x0,x1,y0,y1, s_width, s_height = fn_selection_info(image) # get alpha selection info
                    pdb.gimp_selection_none(image)
                    if selection:
                        if x0<min_x_pos: min_x_pos = x0 # find the minimum x co-ord of all characters
                        if y0<min_y_pos: min_y_pos = y0 # find the minimum y co-ord of all characters
                        if x1>max_x_pos: max_x_pos = x1 # find the maximum x co-ord of all characters
                        if y1>max_y_pos: max_y_pos = y1 # find the maximum y co-ord of all characters
			
                    pdb.gimp_image_remove_layer(image, text_layer) # discard each layer as we go along after gathering data
                loop += 1
		
        if max_y_pos>=font_size*2:
            pdb.gimp_message("There may be a problem with this font. You may want to check the size of the layers and manually crop the image.\n The height of the font was originally set to " + str(font_size) )
	
        loop = 0
		
		
        if len(symbol_list)>1:
		
            loop = 1
            while loop < len(symbol_list): # loop through unicode list to find the max/min positions of characters in all layers
                text_layer = pdb.gimp_text_fontname(image, None,0, 0, eval(symbol_list[loop]), 0, 1, font_size, 1, font_name)
                pdb.gimp_image_select_item(image, 2, text_layer) # character alpha to selection 
                success,x0,x1,y0,y1, s_width, s_height = fn_selection_info(image) # get alpha selection info
                pdb.gimp_selection_none(image)
                if success:
                    #if x0<min_x_pos: min_x_pos = x0 # find the minimum x co-ord of all characters
                    #if y0<min_y_pos: min_y_pos = y0 # find the minimum y co-ord of all characters
                    #if x1>max_x_pos: max_x_pos = x1 # find the maximum x co-ord of all characters
                    #if y1>max_y_pos: max_y_pos = y1 # find the maximum y co-ord of all characters
			
                    temp_vector_layer=pdb.gimp_vectors_new_from_text_layer(image, text_layer)
                    pdb.gimp_image_insert_vectors(image, temp_vector_layer, None, -1)
                    temp_vector_layer = pdb.gimp_item_transform_translate(temp_vector_layer, 0 - x0,0 - y0)
                    temp_layer = gimp.Layer(image, "Temp Layer", image.width, image.height, RGBA_IMAGE, 100, 0) 
                    image.add_layer(temp_layer, -1)
                    pdb.gimp_image_select_item(image, 2, temp_vector_layer)
                    pdb.gimp_drawable_edit_fill(temp_layer, 0)
                    pdb.gimp_selection_none(image)
                    pdb.gimp_image_remove_vectors(image, temp_vector_layer)

                    if name_symbols:
                        new_name = fn_rename_uni_layer(loop,symbol_list[loop])
                        pdb.gimp_item_set_name(temp_layer, new_name)
                    else:
                        pdb.gimp_item_set_name(temp_layer,eval(symbol_list[loop]) + ".png")
						
                    pdb.gimp_image_remove_layer(image, text_layer)
                    if crop_symbols:
                        pdb.gimp_layer_resize(temp_layer, s_width,(y1-y0),0,0)
                    else:
                        pdb.gimp_layer_resize(temp_layer, s_width,(max_y_pos-min_y_pos),0,0)
				
                    if loop <> (len(symbol_list) - 1): 
                        pdb.gimp_item_set_visible(temp_layer, False)
				
                else:
                    pdb.gimp_selection_none(image)
                    if removed == "None":
                        removed = char
                    else:
                        removed += char
                    pdb.gimp_image_remove_layer(image, text_layer) # discard empty text layer				
				
                loop += 1
        else:
		
            while loop < len(layer_string): # loop through each layer to place each character at its new position
                char = layer_string[loop]
                if char <> " ":
                    char = layer_string[loop].decode('utf-8')
                    # create the text layer at the origin of the image
                    text_layer = pdb.gimp_text_fontname(image, None,0, 0, char, 0, 1, font_size, 1, font_name)					
                    pdb.gimp_image_select_item(image, 2, text_layer) # character alpha to selection					
                    empty = pdb.gimp_selection_is_empty(image)
                    if not empty:					
                        selection,x0,x1,y0,y1, s_width, s_height = fn_selection_info(image)
                        success = pdb.gimp_edit_copy(text_layer)
                        pdb.gimp_selection_none(image)
                        if success:
			
                            temp_layer = gimp.Layer(image, "Temp Layer", image.width, image.height, RGBA_IMAGE, 100, 0) 
                            image.add_layer(temp_layer, -1)
                            new_layer = pdb.gimp_edit_paste(temp_layer, False)
                            pdb.gimp_layer_set_offsets(new_layer,0,(y0-min_y_pos) )
                            pdb.gimp_floating_sel_anchor(new_layer)
                    
                            pdb.gimp_item_set_name(temp_layer,char + ".png")
                            pdb.gimp_image_remove_layer(image, text_layer)
		
                            pdb.gimp_layer_resize(temp_layer, s_width,(max_y_pos-min_y_pos),0,0)
                            if loop <> (len(layer_string) - 1): 
                                pdb.gimp_item_set_visible(temp_layer, False)
		
		
                    else:
                        pdb.gimp_selection_none(image)
                        if removed == "None":
                            removed = char
                        else:
                            removed += char
                        pdb.gimp_image_remove_layer(image, text_layer) # discard empty text layer
		
                loop += 1
				
				
        pdb.gimp_image_resize_to_layers(image) # fit the canvas to the new layers
        if len(symbol_list)==1:
            # create a space layer at one third the size of the canvas for non-unicode characters
            temp_layer = gimp.Layer(image, "space.png", int((left_margin+right_margin+max_x_pos)/3), int((max_y_pos-min_y_pos)), RGBA_IMAGE, 100, 0) 
            image.add_layer(temp_layer, -1)
            pdb.gimp_item_set_visible(temp_layer, False)
            layer_names_list = image.layers
            pdb.gimp_image_set_active_layer(image, layer_names_list[1])
	
        if scale_custom:
            current_width = int(max_x_pos) # get current layer height
            current_height = int((max_y_pos-min_y_pos))
            if scale_to_pixel: scale_size = font_size # user wants size to be in pixels rather than Gimp font points
		
            if scale_size<>current_height:
                scale_factor = float((scale_size-(top_margin+bottom_margin))/current_height) # get factor to scale characters with
                scaled_width, scaled_height = int(current_width*scale_factor),int(current_height*scale_factor) # set new width/height
                scaled_width, scaled_height = int(image.width*scale_factor),int(image.height*scale_factor) # set new width/height
                #scaled_width, scaled_height = (current_width*scale_factor),(current_height*scale_factor) # set new width/height
                old_interpolation = pdb.gimp_context_get_interpolation() # Note current interpolation method
                pdb.gimp_context_set_interpolation(3) # Ensure that interpolation is set to No Halo method
                pdb.gimp_image_scale(image, scaled_width, scaled_height)
                pdb.gimp_context_set_interpolation(old_interpolation) # Restore old interpolation method
			
                
                loop = 0
                num_layers, layer_ids = pdb.gimp_image_get_layers(image)
                layer_names_list = image.layers
                while loop < num_layers: # loop through character layers and resize to new size (taken from height) with margins added
                    layer_width = layer_names_list[loop].width
                    layer_height = layer_names_list[loop].height
                    pdb.gimp_layer_resize(layer_names_list[loop],layer_width + left_margin + right_margin, layer_height + top_margin + bottom_margin, left_margin, top_margin)
                    pdb.gimp_layer_set_offsets(layer_names_list[loop],0,0)
                    loop += 1
                pdb.gimp_image_resize_to_layers(image)
                
			
        if removed<>"None":
            removed = removed.decode('utf-8')
            removed_len = len(removed)
            if removed_len == 1:
                my_message = "There is no " + removed + " character available to use with this font. You may wish to add your own layer manually."
            else:
                dummy = 0
                my_message = "There are " + str(removed_len) + " characters missing from this particular font. They are shown below:\n" + removed
            pdb.gimp_message(my_message)
			
    pdb.gimp_displays_flush()

    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
		
######################################################################################################################################
	
def fn_get_unicode_value(uni_input1,uni_input2, symbol_list):
    if uni_input2>0:
        for uni_val in range(uni_input1, uni_input2+1):
            hex_string = fn_unicode_hex(uni_val)
            symbol_list.append(hex_string)
    else:
        hex_string = fn_unicode_hex(uni_input1)
        symbol_list.append(hex_string)
    return symbol_list

######################################################################################################################################

def fn_rename_uni_layer(lp, h_string):
    # rename unicode layer
    try:
        alternative_name = unicodedata.name(eval(h_string)).lower()
    except ValueError:
        alternative_name = "Unicode Layer " + str(lp)
    return alternative_name

######################################################################################################################################

def fn_unicode_hex(dec_number):
    # convert a number into hex value string to use with unicode escape sequence
    hex_lead = "U" + "\'" + "\\" + "U"
    hex_tail = "\'"
    hex_middle = str(hex(dec_number)).upper()
    hex_middle = hex_middle[2:len(hex_middle)]
    zero_string = ""
    if len(hex_middle)<8:
        zero_string = "0" * (8 - len(hex_middle))
    hex_name = hex_lead + zero_string + hex_middle + hex_tail
    return hex_name

######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Layers_With_Forced_Margins_v0_04_Tabbed",
    "Provides options for creating custom font file layers from fonts",
    "Provides options for creating custom font file layers from fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2019",
    "<Image>/Filters/Custom Font Tools/Custom Font Layers From Fonts With Forced Margins v0.04 Tabbed...",
    "RGB*",
    [
    (PF_FONT, "pf_font", "Font: ", "Sans"),
    (PF_SPINNER, "pf_fontsize", "Font size:", 600, (10, 3000, 1)),
    (PF_SPINNER, "leftmarginsize", "Left-hand margin size", 24, (0, 1000, 1)),
    (PF_SPINNER, "topmarginsize", "Top margin size", 24, (0, 1000, 1)),
    (PF_SPINNER, "rightmarginsize", "Right-hand margin size", 24, (0, 1000, 1)),
    (PF_SPINNER, "bottommarginsize", "Bottom margin size", 24, (0, 1000, 1)),
    (PF_TOGGLE, "pf_add_ucase", "Add upper case layers:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_add_lcase", "Add lower case layers:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_add_numbercase", "Add numeral layers:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_add_punc", "Add punctuation characters:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_add_keyboard", "Add other keyboard characters:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_add_specials", "Add special characters:", 1, [("No"),("Yes") ]),
    (PF_TOGGLE, "pf_restrict_set", "Create only the layers below:", 0, [("No"),("Yes") ]),
    (PF_STRING, "pf_string", "Characters to use:", "321gfeCBA"),
    (PF_OPTION, "pf_option_languages", "Characters for non-English languages", 0,("None","Nederlands","Français","Deutsch","Italiano","Polski","Português","Español","All of these"),{'newLine':'1','tab':1}),
    (PF_STRING, "pf_string_add_veryspecials", "Add some unicode symbols:", "None", 0,{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_crop_veryspecials", "Crop unicode character layers:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_name_veryspecials", "Rename unicode layers where possible:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_list_veryspecials", "Display keywords for unicode symbols:", 0, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_scale", "Scale the custom font:", 1, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_scaletopixel", "Scale the custom font to actual pixel size:", 1, [("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_scalesize", "Height of scaled font:", 600, (50, 1000, 1),{'newLine':'1','tab':1}),
    (PF_TAB    , "ptab"    , "tabs", 0, (_('Sizing/Spacing/Characters'),_('Languages/Unicode/Scaling')) ),
    ],
    [],
    do_stuff)

main()
